'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = require('fs-extra');
var fs__default = _interopDefault(fs);
var path = require('path');
var JSON5 = require('json5');
var vscodeTextmate = require('vscode-textmate');
var crypto = require('crypto');
var manager = _interopDefault(require('cache-manager'));
var fsStore = _interopDefault(require('cache-manager-fs-hash'));
var unistUtilSelect = require('unist-util-select');
var memoize = _interopDefault(require('micro-memoize'));
var pAll = _interopDefault(require('p-all'));

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function loadJSONTheme(themePath) {
  var fileContents = fs.readFileSync(themePath, 'utf-8');
  return JSON5.parse(fileContents);
} // function loadPListTheme(themePath: string): IRawTheme {
//   const fileContents = fs.readFileSync(themePath, 'utf-8');
//
//   return plistParse(fileContents);
// }


function toShikiTheme(rawTheme) {
  var shikiTheme = _extends({}, rawTheme, {
    bg: getThemeBg(rawTheme)
  });

  if (rawTheme.include) {
    shikiTheme.include = rawTheme.include;
  }

  if (rawTheme.tokenColors) {
    shikiTheme.settings = rawTheme.tokenColors;
  }

  return shikiTheme;
}
/**
 * @param themePath Absolute path to theme.json / theme.tmTheme
 */


function loadTheme(themePath) {
  var theme;
  theme = loadJSONTheme(themePath); // if (/\.json$/.test(themePath)) {
  //   theme = loadJSONTheme(themePath);
  // } else {
  //   theme = loadPListTheme(themePath);
  // }

  var shikiTheme = toShikiTheme(theme);

  if (shikiTheme.include) {
    var includedThemePath = path.resolve(themePath, '..', shikiTheme.include);
    var includedTheme = loadTheme(includedThemePath);

    if (includedTheme.settings) {
      shikiTheme.settings = shikiTheme.settings.concat(includedTheme.settings);
    }

    if (includedTheme.bg && !shikiTheme.bg) {
      shikiTheme.bg = includedTheme.bg;
    }
  }

  return shikiTheme;
}

function getThemeBg(theme) {
  if (theme.colors && theme.colors['editor.background']) {
    return theme.colors['editor.background'];
  }

  var settings = theme.settings ? theme.settings : theme.tokenColors;
  var globalSetting = settings.find(function (s) {
    return !s.name && !s.scope;
  });
  return globalSetting ? globalSetting.settings.background : null;
}

var vscThemes = ['abyss', 'dark_plus', 'dark_vs', 'hc_black', 'kimbie_dark', 'light_plus', 'light_vs', 'monokai', 'monokai_dimmed', 'quietlight', 'red', 'solarized_dark', 'solarized_light'];
var materialThemes = ['Material-Theme-Darker-High-Contrast', 'Material-Theme-Darker', 'Material-Theme-Default-High-Contrast', 'Material-Theme-Default', 'Material-Theme-Lighter-High-Contrast', 'Material-Theme-Lighter', 'Material-Theme-Ocean-High-Contrast', 'Material-Theme-Ocean', 'Material-Theme-Palenight-High-Contrast', 'Material-Theme-Palenight'];
var niceThemes = ['nord', 'min-light', 'min-dark', 'white', 'white-night', 'zeit'];

function mapF(subdir) {
  return function (n) {
    var p = fs.existsSync(path.resolve(__dirname, "./data/" + subdir + "/" + n + ".json")) ? path.resolve(__dirname, "./data/" + subdir + "/" + n + ".json") : path.resolve(__dirname, "./data/" + subdir + "/" + n + ".tmTheme");
    return loadTheme(p);
  };
}

function getTheme(t) {
  if (vscThemes.includes(t)) {
    return mapF('vscode')(t);
  }

  if (materialThemes.includes(t)) {
    return mapF('material')(t);
  }

  if (niceThemes.includes(t)) {
    return mapF('nice')(t);
  }

  throw Error("No theme " + t + " found");
}

var commonLangIds = ['clojure', 'c', 'cpp', 'csharp', 'css', 'go', 'html', 'jade', 'java', 'jsx', 'javascript', 'json', 'jsonc', 'less', 'lua', 'makefile', 'markdown', 'objective-c', 'perl6', 'php', 'python', 'r', 'ruby', 'rust', 'scss', 'shellscript', 'sql', 'swift', 'typescript', 'tsx', 'xml', 'yaml',
/**
 * Extra grammars
 */
'haml', 'graphql', 'sass', 'stylus', 'postcss', 'vue', 'vue-html', 'latex', 'toml', 'clarity', 'kotlin', 'nginx'];
var commonLangAliases = ['clj', 'c++', 'htm', 'xhtml', 'js', 'objc', 'py', 'rb', 'shell', 'bash', 'sh', 'zsh', 'ts', 'yml', 'md',
/**
 * Extra grammars
 */
'styl', 'tex', 'clar'];
function getLangRegistrations(langs) {
  var langRegistrationMap = {};
  languages.forEach(function (l) {
    // @ts-ignore
    langRegistrationMap[l.id] = l;
    l.aliases.forEach(function (a) {
      // @ts-ignore
      langRegistrationMap[a] = l;
    });
  }); // @ts-ignore

  return langs.map(function (l) {
    return langRegistrationMap[l];
  });
}
var languages = [{
  id: 'bat',
  scopeName: 'source.dosbatch',
  path: './data/grammars/Batch File.tmLanguage',
  aliases: ['batch']
}, {
  id: 'clojure',
  scopeName: 'source.clojure',
  path: './data/grammars/Clojure.tmLanguage',
  aliases: ['clj']
}, {
  id: 'coffeescript',
  scopeName: 'source.coffee',
  path: './data/grammars/coffeescript.json',
  aliases: ['coffee']
}, {
  id: 'c',
  scopeName: 'source.c',
  path: './data/grammars/c.json',
  aliases: []
}, {
  id: 'cpp',
  scopeName: 'source.cpp',
  path: './data/grammars/c++.json',
  aliases: ['c++']
}, {
  id: 'csharp',
  scopeName: 'source.cs',
  path: './data/grammars/csharp.tmLanguage.json',
  aliases: []
}, {
  id: 'css',
  scopeName: 'source.css',
  path: './data/grammars/css.plist',
  aliases: []
}, {
  id: 'diff',
  scopeName: 'source.diff',
  path: './data/grammars/diff.tmLanguage',
  aliases: []
}, {
  id: 'dockerfile',
  scopeName: 'source.dockerfile',
  path: './data/grammars/Dockerfile.tmLanguage',
  aliases: ['docker']
}, {
  id: 'fsharp',
  scopeName: 'source.fsharp',
  path: './data/grammars/fsharp.json',
  aliases: ['f#']
}, {
  id: 'git-commit',
  scopeName: 'text.git-commit',
  path: './data/grammars/git-commit.tmLanguage',
  aliases: []
}, {
  id: 'git-rebase',
  scopeName: 'text.git-rebase',
  path: './data/grammars/git-rebase.tmLanguage',
  aliases: []
}, {
  id: 'go',
  scopeName: 'source.go',
  path: './data/grammars/go.json',
  aliases: []
}, {
  id: 'groovy',
  scopeName: 'source.groovy',
  path: './data/grammars/Groovy.tmLanguage',
  aliases: []
}, {
  id: 'handlebars',
  scopeName: 'text.html.handlebars',
  path: './data/grammars/Handlebars.json',
  aliases: ['hbs']
}, {
  id: 'html',
  scopeName: 'text.html.basic',
  path: './data/grammars/html.json',
  aliases: ['htm', 'xhtml']
}, {
  id: 'ini',
  scopeName: 'source.properties',
  path: './data/grammars/properties.plist',
  aliases: []
}, {
  id: 'properties',
  scopeName: 'source.properties',
  path: './data/grammars/properties.plist',
  aliases: []
}, {
  id: 'jade',
  scopeName: 'text.jade',
  path: './data/grammars/Jade.json',
  aliases: []
}, {
  id: 'java',
  scopeName: 'source.java',
  path: './data/grammars/java.json',
  aliases: []
}, {
  id: 'jsx',
  scopeName: 'source.js',
  path: './data/grammars/JavaScript.tmLanguage.json',
  aliases: []
}, {
  id: 'javascript',
  scopeName: 'source.js',
  path: './data/grammars/JavaScript.tmLanguage.json',
  aliases: ['js']
}, {
  id: 'json',
  scopeName: 'source.json',
  path: './data/grammars/JSON.tmLanguage',
  aliases: []
}, {
  id: 'jsonc',
  scopeName: 'source.json.comments',
  path: './data/grammars/JSONC.tmLanguage.json',
  aliases: []
}, {
  id: 'less',
  scopeName: 'source.css.less',
  path: './data/grammars/less.tmLanguage.json',
  aliases: []
}, {
  id: 'lua',
  scopeName: 'source.lua',
  path: './data/grammars/lua.json',
  aliases: []
}, {
  id: 'makefile',
  scopeName: 'source.makefile',
  path: './data/grammars/Makefile.json',
  aliases: []
}, {
  id: 'markdown',
  scopeName: 'text.html.markdown',
  path: './data/grammars/markdown.tmLanguage',
  aliases: ['md']
}, {
  id: 'objective-c',
  scopeName: 'source.objc',
  path: './data/grammars/Objective-C.tmLanguage',
  aliases: ['objc']
}, {
  id: 'perl',
  scopeName: 'source.perl',
  path: './data/grammars/Perl.plist',
  aliases: []
}, {
  id: 'perl6',
  scopeName: 'source.perl.6',
  path: './data/grammars/Perl 6.tmLanguage',
  aliases: []
}, {
  id: 'php',
  scopeName: 'text.html.php',
  path: './data/grammars/php.json',
  aliases: []
}, {
  id: 'powershell',
  scopeName: 'source.powershell',
  path: './data/grammars/PowershellSyntax.tmLanguage',
  aliases: ['ps', 'ps1']
}, {
  id: 'python',
  scopeName: 'source.python',
  path: './data/grammars/MagicPython.tmLanguage.json',
  aliases: ['py']
}, {
  id: 'r',
  scopeName: 'source.r',
  path: './data/grammars/R.plist',
  aliases: []
}, {
  id: 'razor',
  scopeName: 'text.html.cshtml',
  path: './data/grammars/cshtml.json',
  aliases: []
}, {
  id: 'ruby',
  scopeName: 'source.ruby',
  path: './data/grammars/Ruby.plist',
  aliases: ['rb']
}, {
  id: 'rust',
  scopeName: 'source.rust',
  path: './data/grammars/rust.json',
  aliases: []
}, {
  id: 'scss',
  scopeName: 'source.css.scss',
  path: './data/grammars/scss.json',
  aliases: []
}, {
  id: 'shaderlab',
  scopeName: 'source.shaderlab',
  path: './data/grammars/shaderlab.json',
  aliases: ['shader']
}, {
  id: 'shellscript',
  scopeName: 'source.shell',
  path: './data/grammars/Shell-Unix-Bash.tmLanguage.json',
  aliases: ['shell', 'bash', 'sh', 'zsh']
}, {
  id: 'sql',
  scopeName: 'source.sql',
  path: './data/grammars/SQL.plist',
  aliases: []
}, {
  id: 'swift',
  scopeName: 'source.swift',
  path: './data/grammars/swift.json',
  aliases: []
}, {
  id: 'typescript',
  scopeName: 'source.ts',
  path: './data/grammars/TypeScript.tmLanguage.json',
  aliases: ['ts']
}, {
  id: 'tsx',
  scopeName: 'source.tsx',
  path: './data/grammars/TypeScriptReact.tmLanguage.json',
  aliases: []
}, {
  id: 'vb',
  scopeName: 'source.asp.vb.net',
  path: './data/grammars/ASPVBnet.plist',
  aliases: ['cmd']
}, {
  id: 'xml',
  scopeName: 'text.xml',
  path: './data/grammars/xml.json',
  aliases: []
}, {
  id: 'xsl',
  scopeName: 'text.xml.xsl',
  path: './data/grammars/xsl.json',
  aliases: []
}, {
  id: 'yaml',
  scopeName: 'source.yaml',
  path: './data/grammars/yaml.json',
  aliases: ['yml']
},
/**
 * Extra grammars
 */
{
  id: 'haml',
  scopeName: 'text.haml',
  path: './data/extraGrammars/haml.json',
  aliases: []
}, {
  id: 'graphql',
  scopeName: 'source.graphql',
  path: './data/extraGrammars/graphql.json',
  aliases: []
}, {
  id: 'sass',
  scopeName: 'source.sass',
  path: './data/extraGrammars/sass.tmLanguage',
  aliases: []
}, {
  id: 'stylus',
  scopeName: 'source.stylus',
  path: './data/extraGrammars/stylus.json',
  aliases: ['styl']
}, {
  id: 'postcss',
  scopeName: 'source.css.postcss',
  path: './data/extraGrammars/postcss.json',
  aliases: []
}, {
  id: 'vue',
  scopeName: 'source.vue',
  path: './data/extraGrammars/vue.json',
  aliases: []
}, {
  id: 'vue-html',
  scopeName: 'text.html.vue-html',
  path: './data/extraGrammars/vue-html.json',
  aliases: []
}, {
  id: 'latex',
  scopeName: 'text.tex.latex',
  path: './data/extraGrammars/latex.plist',
  aliases: ['tex']
}, {
  id: 'toml',
  scopeName: 'source.toml',
  path: './data/extraGrammars/TOML.tmLanguage',
  aliases: []
}, {
  id: 'clarity',
  scopeName: 'source.clarity',
  path: './data/extraGrammars/clarity.tmLanguage.json',
  aliases: ['clar']
}, {
  id: 'kotlin',
  scopeName: 'source.kotlin',
  path: './data/extraGrammars/kotlin.tmLanguage',
  aliases: []
}, {
  id: 'nginx',
  scopeName: 'source.nginx',
  path: './data/extraGrammars/nginx.tmLanguage',
  aliases: ['conf']
}];
languages.forEach(function (l) {
  l.path = path.resolve(__dirname, l.path);
});

// A type of promise-like that resolves synchronously and supports only one observer

const _iteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.iterator || (Symbol.iterator = Symbol("Symbol.iterator"))) : "@@iterator";

const _asyncIteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.asyncIterator || (Symbol.asyncIterator = Symbol("Symbol.asyncIterator"))) : "@@asyncIterator";

// Asynchronously call a function and send errors to recovery continuation
function _catch(body, recover) {
	try {
		var result = body();
	} catch(e) {
		return recover(e);
	}
	if (result && result.then) {
		return result.then(void 0, recover);
	}
	return result;
}

/*---------------------------------------------------------
 * Copyright (C) Microsoft Corporation. All rights reserved.
 *--------------------------------------------------------*/

// @ts-ignore
var readGrammarFromPath = function readGrammarFromPath(path, cache) {
  try {
    return Promise.resolve(function () {
      if (cache) {
        return _catch(function () {
          var filename = generateHash(getFilename(path));
          return Promise.resolve(cache.get(filename)).then(function (cached) {
            return cached ? cached : Promise.resolve(readFile(path, 'utf8')).then(function (content) {
              var rawGrammar = vscodeTextmate.parseRawGrammar(content.toString(), path);
              return Promise.resolve(cache.set(filename, rawGrammar)).then(function () {
                return rawGrammar;
              });
            });
          });
        }, function (e) {
          console.log(e);
        });
      } else {
        return Promise.resolve(readFile(path, 'utf8')).then(function (content) {
          return vscodeTextmate.parseRawGrammar(content.toString(), path);
        });
      }
    }());
  } catch (e) {
    return Promise.reject(e);
  }
};

var generateHash = function generateHash(node) {
  return crypto.createHash('md5').update(JSON.stringify(node)).digest('hex');
};

var util = /*#__PURE__*/require('util');

var readFile = /*#__PURE__*/util.promisify(fs.readFile);

function getFilename(path) {
  var filename = path.split('mars/')[1];
  return filename;
}

var Resolver = /*#__PURE__*/function () {
  function Resolver(languages, onigLibPromise, onigLibName, cache) {
    var _this = this;

    this.langMap = {};
    this.scopeToLangMap = {};
    this._languages = languages;
    this._onigLibPromise = onigLibPromise;
    this._onigLibName = onigLibName;
    this.cache = cache;

    this._languages.forEach(function (l) {
      _this.langMap[l.id] = l;
      l.aliases.forEach(function (a) {
        _this.langMap[a] = l;
      });
      _this.scopeToLangMap[l.scopeName] = l;
    });
  }

  var _proto = Resolver.prototype;

  _proto.getOnigLib = function getOnigLib() {
    return this._onigLibPromise;
  };

  _proto.getOnigLibName = function getOnigLibName() {
    return this._onigLibName;
  };

  _proto.getLangRegistration = function getLangRegistration(langIdOrAlias) {
    return this.langMap[langIdOrAlias];
  };

  _proto.loadGrammar = function loadGrammar(scopeName) {
    try {
      var _temp3 = function _temp3(_result) {
        return _exit2 ? _result : Promise.resolve(readGrammarFromPath(lang.path, _this3.cache)).then(function (g) {
          lang.grammar = g;
          return g;
        });
      };

      var _exit2 = false;

      var _this3 = this;

      var lang = _this3.scopeToLangMap[scopeName];

      if (!lang) {
        // @ts-ignore
        return Promise.resolve(null);
      }

      if (lang.grammar) {
        return Promise.resolve(lang.grammar);
      }

      var _temp4 = function () {
        if (_this3.cache) {
          var filename = getFilename(lang.path);
          return Promise.resolve(_this3.cache.get(filename)).then(function (cached) {
            if (cached) {
              _exit2 = true;
              return cached;
            }
          });
        }
      }();

      return Promise.resolve(_temp4 && _temp4.then ? _temp4.then(_temp3) : _temp3(_temp4));
    } catch (e) {
      return Promise.reject(e);
    }
  };

  return Resolver;
}();

/*---------------------------------------------------------
 * Copyright (C) Microsoft Corporation. All rights reserved.
 *--------------------------------------------------------*/

var onigasmLib;
function getOnigasm() {
  if (!onigasmLib) {
    var onigasmModule = require('onigasm');

    var onigasmIndexpath = require.resolve('onigasm');

    var fs = require('fs-extra');

    var path = require('path');

    var wasmBin = fs.readFileSync(path.join(onigasmIndexpath, '../onigasm.wasm')).buffer;
    onigasmLib = onigasmModule.loadWASM(wasmBin).then(function (_) {
      return {
        createOnigScanner: function createOnigScanner(patterns) {
          return new onigasmModule.OnigScanner(patterns);
        },
        createOnigString: function createOnigString(s) {
          return new onigasmModule.OnigString(s);
        }
      };
    });
  }

  return onigasmLib;
}

var StackElementMetadata = /*#__PURE__*/function () {
  function StackElementMetadata() {}

  StackElementMetadata.toBinaryStr = function toBinaryStr(metadata) {
    var r = metadata.toString(2);

    while (r.length < 32) {
      r = '0' + r;
    }

    return r;
  };

  StackElementMetadata.printMetadata = function printMetadata(metadata) {
    var languageId = StackElementMetadata.getLanguageId(metadata);
    var tokenType = StackElementMetadata.getTokenType(metadata);
    var fontStyle = StackElementMetadata.getFontStyle(metadata);
    var foreground = StackElementMetadata.getForeground(metadata);
    var background = StackElementMetadata.getBackground(metadata);
    console.log({
      languageId: languageId,
      tokenType: tokenType,
      fontStyle: fontStyle,
      foreground: foreground,
      background: background
    });
  };

  StackElementMetadata.getLanguageId = function getLanguageId(metadata) {
    return (metadata & 255
    /* LANGUAGEID_MASK */
    ) >>> 0
    /* LANGUAGEID_OFFSET */
    ;
  };

  StackElementMetadata.getTokenType = function getTokenType(metadata) {
    return (metadata & 1792
    /* TOKEN_TYPE_MASK */
    ) >>> 8
    /* TOKEN_TYPE_OFFSET */
    ;
  };

  StackElementMetadata.getFontStyle = function getFontStyle(metadata) {
    return (metadata & 14336
    /* FONT_STYLE_MASK */
    ) >>> 11
    /* FONT_STYLE_OFFSET */
    ;
  };

  StackElementMetadata.getForeground = function getForeground(metadata) {
    return (metadata & 8372224
    /* FOREGROUND_MASK */
    ) >>> 14
    /* FOREGROUND_OFFSET */
    ;
  };

  StackElementMetadata.getBackground = function getBackground(metadata) {
    return (metadata & 4286578688
    /* BACKGROUND_MASK */
    ) >>> 23
    /* BACKGROUND_OFFSET */
    ;
  };

  StackElementMetadata.set = function set(metadata, languageId, tokenType, fontStyle, foreground, background) {
    var _languageId = StackElementMetadata.getLanguageId(metadata);

    var _tokenType = StackElementMetadata.getTokenType(metadata);

    var _fontStyle = StackElementMetadata.getFontStyle(metadata);

    var _foreground = StackElementMetadata.getForeground(metadata);

    var _background = StackElementMetadata.getBackground(metadata);

    if (languageId !== 0) {
      _languageId = languageId;
    }

    if (tokenType !== 0
    /* Other */
    ) {
        _tokenType = tokenType === 8
        /* MetaEmbedded */
        ? 0
        /* Other */
        : tokenType;
      }

    if (fontStyle !== -1
    /* NotSet */
    ) {
        _fontStyle = fontStyle;
      }

    if (foreground !== 0) {
      _foreground = foreground;
    }

    if (background !== 0) {
      _background = background;
    }

    return (_languageId << 0
    /* LANGUAGEID_OFFSET */
    | _tokenType << 8
    /* TOKEN_TYPE_OFFSET */
    | _fontStyle << 11
    /* FONT_STYLE_OFFSET */
    | _foreground << 14
    /* FOREGROUND_OFFSET */
    | _background << 23
    /* BACKGROUND_OFFSET */
    ) >>> 0;
  };

  return StackElementMetadata;
}();

/*---------------------------------------------------------
 * Copyright (C) Microsoft Corporation. All rights reserved.
 *--------------------------------------------------------*/
function tokenizeWithTheme(theme, colorMap, fileContents, grammar) {
  var lines = fileContents.split(/\r\n|\r|\n/);
  var ruleStack = null;
  var actual = [];
  var _final = [];

  for (var i = 0, len = lines.length; i < len; i++) {
    var line = lines[i];

    if (line === '') {
      actual = [];

      _final.push([]);

      continue;
    } // @ts-ignore


    var resultWithScopes = grammar.tokenizeLine(line, ruleStack);
    var tokensWithScopes = resultWithScopes.tokens; // @ts-ignore

    var result = grammar.tokenizeLine2(line, ruleStack);
    var tokensLength = result.tokens.length / 2;
    var tokensWithScopesIndex = 0;

    for (var j = 0; j < tokensLength; j++) {
      var startIndex = result.tokens[2 * j];
      var nextStartIndex = j + 1 < tokensLength ? result.tokens[2 * j + 2] : line.length;
      var tokenText = line.substring(startIndex, nextStartIndex);

      if (tokenText === '') {
        continue;
      }

      var metadata = result.tokens[2 * j + 1];
      var foreground = StackElementMetadata.getForeground(metadata);
      var foregroundColor = colorMap[foreground];
      var explanation = [];
      var tmpTokenText = tokenText;

      while (tmpTokenText.length > 0) {
        var tokenWithScopes = tokensWithScopes[tokensWithScopesIndex];
        var tokenWithScopesText = line.substring(tokenWithScopes.startIndex, tokenWithScopes.endIndex);
        tmpTokenText = tmpTokenText.substring(tokenWithScopesText.length);
        explanation.push({
          content: tokenWithScopesText,
          scopes: explainThemeScopes(theme, tokenWithScopes.scopes)
        });
        tokensWithScopesIndex++;
      }

      actual.push({
        content: tokenText,
        color: foregroundColor,
        explanation: explanation
      });
    }

    _final.push(actual);

    actual = [];
    ruleStack = result.ruleStack;
  }

  return _final;
}

function explainThemeScopes(theme, scopes) {
  var result = [];

  for (var i = 0, len = scopes.length; i < len; i++) {
    var parentScopes = scopes.slice(0, i);
    var scope = scopes[i];
    result[i] = {
      scopeName: scope,
      themeMatches: explainThemeScope(theme, scope, parentScopes)
    };
  }

  return result;
}

function matchesOne(selector, scope) {
  var selectorPrefix = selector + '.';

  if (selector === scope || scope.substring(0, selectorPrefix.length) === selectorPrefix) {
    return true;
  }

  return false;
}

function matches(selector, selectorParentScopes, scope, parentScopes) {
  if (!matchesOne(selector, scope)) {
    return false;
  }

  var selectorParentIndex = selectorParentScopes.length - 1;
  var parentIndex = parentScopes.length - 1;

  while (selectorParentIndex >= 0 && parentIndex >= 0) {
    if (matchesOne(selectorParentScopes[selectorParentIndex], parentScopes[parentIndex])) {
      selectorParentIndex--;
    }

    parentIndex--;
  }

  if (selectorParentIndex === -1) {
    return true;
  }

  return false;
}

function explainThemeScope(theme, scope, parentScopes) {
  var result = [],
      resultLen = 0;

  for (var i = 0, len = theme.settings.length; i < len; i++) {
    var setting = theme.settings[i];
    var selectors = void 0;

    if (typeof setting.scope === 'string') {
      selectors = setting.scope.split(/,/).map(function (scope) {
        return scope.trim();
      });
    } else if (Array.isArray(setting.scope)) {
      selectors = setting.scope;
    } else {
      continue;
    }

    for (var j = 0, lenJ = selectors.length; j < lenJ; j++) {
      var rawSelector = selectors[j];
      var rawSelectorPieces = rawSelector.split(/ /);
      var selector = rawSelectorPieces[rawSelectorPieces.length - 1];
      var selectorParentScopes = rawSelectorPieces.slice(0, rawSelectorPieces.length - 1);

      if (matches(selector, selectorParentScopes, scope, parentScopes)) {
        // match!
        result[resultLen++] = setting; // break the loop

        j = lenJ;
      }
    }
  }

  return result;
}

function renderToHtml(lines, options) {
  if (options === void 0) {
    options = {};
  }

  var bg = options.bg || '#fff';
  var html = '';
  html += "<pre class=\"shiki\" style=\"background-color: " + bg + "\">";

  if (options.langId) {
    html += "<div class=\"language-id\">" + options.langId + "</div>";
  }

  html += "<code>";
  lines.forEach(function (l) {
    if (l.length === 0) {
      html += "\n";
    } else {
      l.forEach(function (token) {
        html += "<span style=\"color: " + token.color + "\">" + escapeHtml(token.content) + "</span>";
      });
      html += "\n";
    }
  });
  html = html.replace(/\n*$/, ''); // Get rid of final new lines

  html += "</code></pre>";
  return html;
}

function escapeHtml(html) {
  return html.replace(/</g, '&lt;').replace(/>/g, '&gt;');
}

var getHighlighter = function getHighlighter(options) {
  try {
    if (highlighter) {
      return Promise.resolve(highlighter);
    }

    var t;

    if (typeof options.theme === 'string') {
      t = getTheme(options.theme);
    } else if (options.theme.name) {
      t = options.theme;
    } else {
      t = getTheme('nord');
    }

    var langs = [].concat(commonLangIds, commonLangAliases);
    var renderer = renderToHtml;

    if (options.langs) {
      langs = options.langs;
    }

    if (options.renderer) {
      renderer = options.renderer;
    }

    var langRegistrations = getLangRegistrations(langs);
    var s = new Shiki(t, langRegistrations, renderer, options.cache);
    return Promise.resolve(s.getHighlighter()).then(function (_s$getHighlighter) {
      highlighter = _s$getHighlighter;
      return highlighter;
    });
  } catch (e) {
    return Promise.reject(e);
  }
};
var highlighter;
var languageRecord = {};

var Shiki = /*#__PURE__*/function () {
  function Shiki(theme, langs, renderer, cache) {
    this._resolver = new Resolver(langs, getOnigasm(), 'onigasm', cache);
    this._registry = new vscodeTextmate.Registry(this._resolver);
    this._renderer = renderer;
    this._cache = cache;

    this._registry.setTheme(theme);

    this._theme = theme;
    this._colorMap = this._registry.getColorMap();
    this._langs = langs;
  }

  var _proto = Shiki.prototype;

  _proto.getGrammar = function getGrammar() {
    try {
      var _temp4 = function _temp4(_result) {
        return _exit2 ? _result : Promise.resolve(Promise.all( // @ts-ignore
        _this2._langs.map(function (l) {
          try {
            return Promise.resolve(_this2._registry.loadGrammar(l.scopeName)).then(function (g) {
              // @ts-ignore
              languageRecord[l.id] = g;
              l.aliases.forEach(function (la) {
                // @ts-ignore
                languageRecord[la] = g;
              });
            });
          } catch (e) {
            return Promise.reject(e);
          }
        }))).then(function () {
          var _temp = function () {
            if (_this2._cache) {
              return Promise.resolve(_this2._cache.set('languageRecord', languageRecord)).then(function () {});
            }
          }();

          return _temp && _temp.then ? _temp.then(function () {
            return languageRecord;
          }) : languageRecord;
        });
      };

      var _exit2 = false;

      var _this2 = this;

      var _temp5 = function () {
        if (_this2._cache) {
          return Promise.resolve(_this2._cache.get('languageRecord')).then(function (cached) {
            if (cached) {
              _exit2 = true;
              return cached;
            }
          });
        }
      }();

      return Promise.resolve(_temp5 && _temp5.then ? _temp5.then(_temp4) : _temp4(_temp5));
    } catch (e) {
      return Promise.reject(e);
    }
  } // @ts-ignore
  ;

  _proto.getHighlighter = function getHighlighter() {
    try {
      var _this4 = this;

      return Promise.resolve(_this4.getGrammar()).then(function (ltog) {
        return {
          codeToThemedTokens: function codeToThemedTokens(code, lang) {
            if (isPlaintext(lang)) {
              throw Error('Cannot tokenize plaintext');
            }

            if (!ltog[lang]) {
              throw Error("No language registration for " + lang);
            }

            return tokenizeWithTheme(_this4._theme, _this4._colorMap, code, ltog[lang]);
          },
          codeToHtml: function codeToHtml(node, props) {
            var lang = node.lang;
            var code = node.value;
            var highlight = props && props.highlight;

            if (isPlaintext(lang)) {
              return _this4._renderer([[{
                content: code
              }]], {
                bg: _this4._theme.bg
              });
            }

            if (!ltog[lang]) {
              console.error("No language registration for " + lang);
              return _this4._renderer([[{
                content: code
              }]], {
                bg: _this4._theme.bg
              });
            }

            var tokens = tokenizeWithTheme(_this4._theme, _this4._colorMap, code, ltog[lang]);
            return _this4._renderer(tokens, {
              bg: _this4._theme.bg,
              highlight: highlight
            });
          }
        };
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  return Shiki;
}();

function isPlaintext(lang) {
  return ['plaintext', 'txt', 'text'].indexOf(lang) !== -1;
}

var MAX_CACHE_SIZE = 250;
var TTL = Number.MAX_SAFE_INTEGER;

var Cache = /*#__PURE__*/function () {
  function Cache(_temp) {
    var _ref = _temp === void 0 ? {} : _temp,
        _ref$name = _ref.name,
        name = _ref$name === void 0 ? "db" : _ref$name,
        _ref$store = _ref.store,
        store = _ref$store === void 0 ? fsStore : _ref$store;

    this.name = name;
    this.store = store;
  } // @ts-ignore


  var _proto = Cache.prototype;

  _proto.init = function init() {
    fs__default.ensureDirSync(this.directory);
    var caches = [{
      store: "memory",
      max: MAX_CACHE_SIZE
    }, {
      store: this.store,
      options: {
        path: this.directory,
        ttl: TTL
      }
    }].map(function (cache) {
      return manager.caching(cache);
    });
    this.cache = manager.multiCaching(caches);
    return this;
  };

  _proto.get = function get(key) {
    var _this = this;

    return new Promise(function (resolve) {
      _this.cache.get(key, function (err, res) {
        resolve(err ? undefined : res);
      });
    });
  };

  _proto.set = function set(key, value, args) {
    var _this2 = this;

    if (args === void 0) {
      args = {};
    }

    return new Promise(function (resolve) {
      _this2.cache.set(key, value, args, function (err) {
        resolve(err ? undefined : value);
      });
    });
  };

  _proto.wrap = function wrap(key, cb) {
    return this.cache.wrap(key, cb);
  };

  _createClass(Cache, [{
    key: "directory",
    get: function get() {
      return path.join(process.cwd(), ".cache/caches/" + this.name);
    }
  }]);

  return Cache;
}();

var cache = /*#__PURE__*/new Cache({
  name: 'remark-shiki'
}).init();
var generateHash$1 = function generateHash(node) {
  return crypto.createHash('md5').update(JSON.stringify(node)).digest('hex');
};
function escapeHtml$1(html) {
  return html.replace(/</g, '&lt;').replace(/>/g, '&gt;');
}
function codeLanguage(node) {
  if (node.lang) return node.lang;
  return 'bash';
}
function renderToHtml$1(lines, options) {
  if (options === void 0) {
    options = {};
  }

  var highlight = options && options.highlight;

  var addHighlightClass = function addHighlightClass(lineNumber) {
    return highlight && highlight.find(function (number) {
      return number === lineNumber + 1;
    }) ? ' token-line--highlighted' : '';
  };

  var html = '';
  html += "<pre className=\"shiki-code\">";

  if (options.langId) {
    html += "<div className=\"language-id\">" + options.langId + "</div>";
  }

  html += "<code>";
  lines.forEach(function (l, index) {
    if (l.length === 0) {
      html += "<span className=\"token-line token-line--empty" + addHighlightClass(index) + "\"></span>";
    } else {
      html += "<span className=\"token-line" + addHighlightClass(index) + "\">";
      l.forEach(function (token) {
        html += "<span style=\"color: " + token.color + "\">" + escapeHtml$1(token.content) + "</span>";
      });
      html += "</span>";
    }
  });
  html = html.replace(/<span className="token-line token-line--empty"><\/span>*$/, ''); // Get rid of final empty lines

  html += "</code></pre>";
  return html;
}

var getHighlightLineNumbers = function getHighlightLineNumbers(str) {
  if (!str) return;
  var numbers = undefined;
  numbers = str.split(',').flatMap(function (s) {
    if (!s.includes('-')) return +s;

    var _s$split = s.split('-'),
        min = _s$split[0],
        max = _s$split[1]; // @ts-ignore


    var _final = Array.from({
      length: max - min + 1
    }, function (_, n) {
      return n + +min;
    });

    return _final;
  });
  return numbers;
};

var getProps = function getProps(meta) {
  var props = meta && meta.split(' ').reduce(function (acc, cur) {
    if (cur && cur.includes('=')) {
      var t = cur.split('=');
      acc[t[0]] = t[1];
      return acc;
    }

    acc[cur] = true;
    return acc;
  }, {});

  if (props.highlight) {
    props.highlight = getHighlightLineNumbers(props.highlight);
  }

  return props;
};

var DEFAULT_THEME = 'Material-Theme-Darker-High-Contrast';
var generateHighlighter = /*#__PURE__*/memoize(function () {
  try {
    return Promise.resolve(getHighlighter({
      theme: DEFAULT_THEME,
      renderer: renderToHtml$1,
      cache: cache
    }));
  } catch (e) {
    return Promise.reject(e);
  }
}, {
  isPromise: true
}); // @ts-ignore

var highlight = function highlight(node) {
  try {
    var hash = generateHash$1(node.value);
    return Promise.resolve(cache.wrap(hash, function () {
      try {
        var props = node && node.meta && getProps(node.meta);
        return Promise.resolve(generateHighlighter()).then(function (highlighter) {
          // @ts-ignore
          return highlighter.codeToHtml(node, props);
        });
      } catch (e) {
        return Promise.reject(e);
      }
    }));
  } catch (e) {
    return Promise.reject(e);
  }
};

var visitor = /*#__PURE__*/memoize(function (node) {
  try {
    node.type = 'html';
    node.children = undefined;
    node.lang = codeLanguage(node);
    return Promise.resolve(highlight(node)).then(function (_highlight) {
      node.value = _highlight;
    });
  } catch (e) {
    return Promise.reject(e);
  }
}, {
  isPromise: true
});

var transformer = function transformer(tree) {
  try {
    var nodes = unistUtilSelect.selectAll('code', tree);
    return Promise.resolve(pAll(nodes.map(function (node) {
      return function () {
        return visitor(node);
      };
    }), {
      concurrency: 25
    })).then(function () {
      return tree;
    });
  } catch (e) {
    return Promise.reject(e);
  }
};

function attacher() {
  return transformer;
}

module.exports = /*#__PURE__*/memoize(attacher);
//# sourceMappingURL=remark-vscode.cjs.development.js.map
