import { IThemedToken } from './themedTokenizer';

export interface HtmlRendererOptions {
  langId?: string;
  bg?: string;
  highlight?: number[];
}

export function renderToHtml(lines: IThemedToken[][], options: HtmlRendererOptions = {}) {
  const bg = options.bg || '#fff';

  let html = '';

  html += `<pre class="shiki" style="background-color: ${bg}">`;
  if (options.langId) {
    html += `<div class="language-id">${options.langId}</div>`;
  }
  html += `<code>`;

  lines.forEach((l: any[]) => {
    if (l.length === 0) {
      html += `\n`;
    } else {
      l.forEach(token => {
        html += `<span style="color: ${token.color}">${escapeHtml(token.content)}</span>`;
      });
      html += `\n`;
    }
  });
  html = html.replace(/\n*$/, ''); // Get rid of final new lines
  html += `</code></pre>`;

  return html;
}

function escapeHtml(html: string) {
  return html.replace(/</g, '&lt;').replace(/>/g, '&gt;');
}
